from __future__ import annotations

import dataclasses
import types as tys
import typing as ty
from contextlib import suppress
from datetime import date, datetime, time
from enum import Enum
from functools import wraps
from inspect import isclass

from .operators import (
    AddOperator,
    CmpOperator,
    ContainedCmpOperator,
    DivideOperator,
    EqCmpOperator,
    GeCmpOperator,
    GtCmpOperator,
    InvertedField,
    IsNullCmpOperator,
    LeCmpOperator,
    LtCmpOperator,
    MathOperator,
    MultiplyOperator,
    NeCmpOperator,
    SubOperator,
)
from .tools import convert_type, watch_changes

LT_GT_SUPPORTED = {int, float, datetime, date, time}
MATH_SUPPORTED = {int, float}


def field_operator[T, **P, RT: CmpOperator](
    func: ty.Callable[ty.Concatenate[Field[T], P], ty.Type[RT]],
) -> ty.Callable[ty.Concatenate[Field[T], P], RT]:
    """
    Decorator for `Field` comparing operators.
    Check operator is available for this field.
    Returns suitable instance of `Operator` class.
    """
    naming_map = {
        "__eq__": "eq",
        "__ne__": "eq",
        "contained": "eq",
        "is_null": "eq",
        "__lt__": "lt_gt",
        "__le__": "lt_gt",
        "__gt__": "lt_gt",
        "__ge__": "lt_gt",
    }

    @wraps(func)
    def _wrapper(self: Field[T], *args: P.args, **kwargs: P.kwargs) -> RT:
        other = args[0] if args else kwargs.get("other")
        op_name = naming_map[func.__name__]
        if not self.inited:
            raise RuntimeError(f"Model '{self.model_name}' is not initialized")
        if not getattr(self, op_name):
            raise TypeError(
                f"{self!r} does not support {func.__name__} operator"
            )
        elif (op := func(self, *args, **kwargs)) not in {
            IsNullCmpOperator,
            ContainedCmpOperator,
        }:
            if not self.compare_type(type(other)):
                raise TypeError(f"unable to compare {self!r} and {other!r}")
        return op(str(self), other)

    return _wrapper


def math_operator[T, **P, RT: MathOperator](
    func: ty.Callable[ty.Concatenate[Field[T], P], ty.Type[RT]],
) -> ty.Callable[ty.Concatenate[Field[T], P], RT]:
    """
    Decorator for `Field` math operators.
    Check operator is available for this field.
    Returns suitable instance of `MathOperator` class.
    """

    @wraps(func)
    def _wrapper(self: Field[T], *args: P.args, **kwargs: P.kwargs) -> RT:
        other = args[0] if args else kwargs.get("other")
        if self.python_type not in MATH_SUPPORTED:
            raise TypeError(
                f"{self!r} does not support {func.__name__} operator"
            )
        if not self.compare_type(type(other)):
            raise TypeError(
                f"unable to execute math operation on {self!r} and {other!r}"
            )
        return func(self, *args, **kwargs)(other)

    return _wrapper


class Field[T]:
    def __init__(self, default_value: T):
        self.default_value = default_value
        self._first = True

        self.inited: bool = False
        self.model_name: str = ""
        self.name: str = ""
        self.python_type: ty.Type[T] | UnionType[T] | None = None
        self.unique: bool | None = None
        self.sql_type: SpecificSQLType | None = None
        self.index: Index | None = None
        self.eq: bool | None = None
        self.lt_gt: bool | None = None

    def __set_name__(self, owner: type, name: str):
        self.model_name = owner.__name__
        self.name = name

    def __get__(self, obj: object | None, owner: type):
        if self._first:
            self._first = False
            return self.default_value
        if obj is None:
            return self
        return obj.__dict__[self.name]

    def __set__(self, obj, value: T):
        obj.__dict__[self.name] = value

    def init(
        self,
        model_name: str,
        name: str,
        python_type: ty.Type[T] | UnionType[T],
        unique: bool = False,
        sql_type: SpecificSQLType | None = None,
        index: Index | None = None,
        eq: bool = True,
        lt_gt: bool = False,
    ) -> None:
        """
        Initialize `Field` instance.

        :param model_name: Name of the model.
        :param name: Name of the field.
        :param python_type: Type of the field.
        :param default_value: Default value of this field.
        :param unique: Is this field unique?
        :param sql_type: SQL type of this field.
            None - type will be inferred from python_type
        :param index: Index of this field.
        :param eq: Are equation operators available for this field?
        :param lt_gt: Are comparing operators available for this field?
        """
        self.model_name = model_name
        self.name = name
        self.python_type = python_type
        self.unique = unique
        self.sql_type = sql_type
        self.index = index
        self.eq = eq
        self.lt_gt = lt_gt
        self.inited = True

    def compare_type(self, type_: ty.Any) -> bool:
        """
        Checks whether the field can be the specified type.
        """
        if isinstance(self.python_type, UnionType):
            return self.python_type.is_contains_type(type_)
        return type_ is self.python_type

    @field_operator
    def __eq__(self, other: T) -> ty.Type[EqCmpOperator]:
        return EqCmpOperator

    @field_operator
    def __ne__(self, other: T) -> ty.Type[NeCmpOperator]:
        return NeCmpOperator

    @field_operator
    def __lt__(self, other: T) -> ty.Type[LtCmpOperator]:
        return LtCmpOperator

    @field_operator
    def __le__(self, other: T) -> ty.Type[LeCmpOperator]:
        return LeCmpOperator

    @field_operator
    def __gt__(self, other: T) -> ty.Type[GtCmpOperator]:
        return GtCmpOperator

    @field_operator
    def __ge__(self, other: T) -> ty.Type[GeCmpOperator]:
        return GeCmpOperator

    @field_operator
    def contained(
        self, sequence: list[T] | tuple[T, ...]
    ) -> ty.Type[ContainedCmpOperator]:
        for el in sequence:
            if not self.compare_type(type(el)):
                raise TypeError(f"unable to compare {self!r} and `{el!r}`")
        return ContainedCmpOperator

    @field_operator
    def is_null(self) -> ty.Type[IsNullCmpOperator]:
        return IsNullCmpOperator

    def __invert__(self) -> InvertedField:
        return InvertedField(str(self))

    @math_operator
    def __add__(self, other: T) -> ty.Type[AddOperator]:
        return AddOperator

    @math_operator
    def __sub__(self, other: T) -> ty.Type[SubOperator]:
        return SubOperator

    @math_operator
    def __mul__(self, other: T) -> ty.Type[MultiplyOperator]:
        return MultiplyOperator

    @math_operator
    def __div__(self, other: T) -> ty.Type[DivideOperator]:
        return DivideOperator

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return f'"{self.model_name}".{self.name}'

    def __repr__(self):
        type_name = (
            self.python_type.__name__
            if isclass(self.python_type)
            else repr(self.python_type)
        )
        return f"<Field {self}:{type_name}>"


@dataclasses.dataclass
class ModelSignature:
    """
    Signature of a model.
    Contains name and information about fields.
    """

    name: str
    fields: list[Field] = dataclasses.field(default_factory=list)


class UnionType[T: ty.Any]:
    """
    type of field that can take different data types.
    """

    def __init__(self, *types: ty.Type[T]):
        self.types = list(types)
        self.nullable = tys.NoneType in self.types
        if self.nullable:
            self.types.remove(tys.NoneType)  # type: ignore

    def __call__(self, obj: ty.Any) -> T:
        """
        Wrapping obj to suitable data type.
        """
        for type_ in self.types:
            with suppress(ValueError, TypeError):
                return convert_type(obj, type_)
        raise ValueError(f"`{obj}` cant be {self}")

    def is_contains_type(self, type_: ty.Type) -> bool:
        return type_ in self.types or (self.nullable and type_ is tys.NoneType)

    def __repr__(self) -> str:
        return (
            f"{'|'.join(map(lambda x: x.__name__, self.types))}"
            f"{'|None' if self.nullable else ''}"
        )


def prepare_model(model: ty.Type[ty.Any]) -> ModelSignature:
    """
    Prepares model to work in db context.
    Wraps model into `watch_changes`.
    Replaces class attributes onto `ModelField` instances.
    :param model: dataclass.
    :returns: signature of model.
    """
    if hasattr(model, "__aiodbc__"):
        return getattr(model, "__aiodbc__")
    if not dataclasses.is_dataclass(model):
        raise TypeError(f"Model `{model.__name__}` is not a dataclass")
    watch_changes(model)
    signature = ModelSignature(model_name := model.__name__)
    for field_name, field_type in ty.get_type_hints(
        model, include_extras=True
    ).items():
        unique = False
        sql_type = index = None
        if ty.get_origin(field_type) is ty.Annotated:
            metadata = field_type.__metadata__
            unique = FieldMod.UNIQUE in metadata
            sql_type = next(
                filter(lambda x: isinstance(x, SpecificSQLType), metadata), None
            )
            index = next(
                filter(lambda x: isinstance(x, Index) or x is Index, metadata),
                None,
            )
            if index is Index:
                index = Index(f"{field_name}_idx")
            field_type = ty.get_args(field_type)[0]
        if ty.get_origin(field_type) is Field:
            field = getattr(model, field_name, None)
            if not isinstance(field, Field):
                field = Field(getattr(model, field_name, None))
            field_type = ty.get_args(field_type)[0]
        else:
            field = Field(getattr(model, field_name, None))
        lt_gt = field_type in LT_GT_SUPPORTED
        if ty.get_origin(field_type) in {ty.Union, tys.UnionType}:
            field_type = UnionType(*ty.get_args(field_type))
            lt_gt = any(field_type.is_contains_type(x) for x in LT_GT_SUPPORTED)
        field.init(
            model_name=model_name,
            name=field_name,
            python_type=field_type,
            unique=unique,
            sql_type=sql_type,
            index=index,
            eq=True,
            lt_gt=lt_gt,
        )
        signature.fields.append(field)
        setattr(model, field_name, field)
        getattr(model, field_name)
    if (id_field := signature.fields[0]).name != "id":
        raise AttributeError(
            f"The first attribute of the model `{model.__name__}` should be `id`"
        )
    if not id_field.compare_type(int):
        raise TypeError(
            f"`id` attribute of model `{model.__name__}` should contains `int`"
        )

    setattr(model, "__aiodbc__", signature)
    return signature


class FieldMod(Enum):
    UNIQUE = "unique"


class SpecificSQLType(str):
    pass


class Index:
    def __init__(self, name: str, unique: bool = False):
        self.name = name
        self.unique = unique



from __future__ import annotations

import typing as ty

if ty.TYPE_CHECKING:
    from .joins import InnerJoin, Join, LeftJoin, RightJoin
    from .models import Field, ModelSignature
    from .operators import InvertedField, MathOperator, Operator
    from .providers import BaseProvider


class AsyncDBCore[Models]:
    signatures: dict[str, ModelSignature] = ...
    dbs: dict[str, BaseProvider] = ...
    db_names: dict[str, str] = ...

    @classmethod
    def init(
        cls, database_path: str, db_name: str = "main", **connection_kwargs
    ) -> None: ...
    @classmethod
    async def close_connections(cls) -> None: ...
    def __init__(self, db: str = "main"):
        self.provider: BaseProvider = ...

    async def execute(self, query: str, args: ty.Sequence[ty.Any] = ()): ...
    async def create_tables(self) -> None: ...
    @ty.overload
    async def insert[Model](self, obj: Model, /) -> Model: ...
    @ty.overload
    async def insert[Model](self, objs: list[Model], /) -> list[Model]: ...
    def _prepare_select_query(
        self,
        model_name: str,
        join: Join[Models] | None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        reverse: bool = False,
        limit: int | None = None,
        offset: int = 0,
    ) -> str: ...
    @ty.overload
    def _convert_data[Model](
        self, model: ty.Type[Model], data: tuple[ty.Any, ...], join: None = None
    ) -> Model: ...
    @ty.overload
    def _convert_data[Model, JoinModel](
        self,
        model: ty.Type[Model],
        data: tuple[ty.Any, ...],
        join: InnerJoin[JoinModel],
    ) -> tuple[Model, JoinModel]: ...
    @ty.overload
    def _convert_data[Model, JoinModel](
        self,
        model: ty.Type[Model],
        data: tuple[ty.Any, ...],
        join: LeftJoin[JoinModel],
    ) -> tuple[Model, JoinModel | None]: ...
    @ty.overload
    def _convert_data[Model, JoinModel](
        self,
        model: ty.Type[Model],
        data: tuple[ty.Any, ...],
        join: RightJoin[JoinModel],
    ) -> tuple[Model | None, JoinModel]: ...
    @ty.overload
    async def fetchone[Model](
        self,
        model: ty.Type[Model],
        *,
        join: None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> Model | None: ...
    @ty.overload
    async def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: InnerJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model, JoinModel] | None: ...
    @ty.overload
    async def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: LeftJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model, JoinModel | None] | None: ...
    @ty.overload
    async def fetchone[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: RightJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> tuple[Model | None, JoinModel] | None: ...
    @ty.overload
    async def fetchall[Model](
        self,
        model: ty.Type[Model],
        *,
        join: None = None,
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[Model]: ...
    @ty.overload
    async def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: InnerJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model, JoinModel]]: ...
    @ty.overload
    async def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: LeftJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model, JoinModel | None]]: ...
    @ty.overload
    async def fetchall[Model, JoinModel](
        self,
        model: ty.Type[Model],
        *,
        join: RightJoin[JoinModel],
        where: Operator | None = None,
        order_by: (
            Field | InvertedField | tuple[Field | InvertedField, ...] | None
        ) = None,
        limit: int | None = None,
        offset: int = 0,
    ) -> list[tuple[Model | None, JoinModel]]: ...
    async def save(self, obj: Models, /) -> None: ...
    async def update[T](
        self,
        model: ty.Type[Models],
        fields: dict[Field[T], T | MathOperator[T]],
        *,
        where: Operator | None = None,
    ) -> None: ...
    async def delete(
        self, model: ty.Type[Models], *, where: Operator | None
    ) -> None: ...
    async def drop_table(self, model: ty.Type[Models], /) -> None: ...



import dataclasses
import hashlib
import typing as ty
from datetime import date, datetime, time


def watch_changes(_cls: ty.Type | None = None, hash_func=hashlib.md5):
    """
    watch_changes decorator allows you to detect
    changes in the attributes of an instances of class.

    It provides by saving hashes of attributes when instance created.
    Note: hashes will get from the string value of the attribute
        `hash_func(str(value).encode()).hexdigest()`

    >>> from dataclasses import dataclass, field
    >>> @watch_changes
    >>> @dataclass
    >>> class A:
    ...     a: int
    ...     b: dict
    >>> a = A(0, {"a": 0})
    >>> has_changes(a)
    False
    >>> a.a = 1
    >>> has_changes(a)
    True
    >>> a.a = 0
    >>> has_changes(a)
    False
    >>> a.b["a"] += 1
    >>> has_changes(a)
    True
    """

    def _decorator(cls):
        def init_hashes(obj: ty.Any, *args, **kwargs) -> None:
            getattr(obj, "__old_init__")(*args, **kwargs)
            hf = getattr(obj, "__wc_hash_func")
            hashes: dict[str, str] = {}
            for varname, value in obj.__dict__.items():
                if not varname.startswith("_"):
                    hashes[varname] = hf(str(value).encode()).hexdigest()
            setattr(obj, "__wc_hashes", hashes)

        setattr(cls, "__old_init__", cls.__init__)
        setattr(cls, "__wc_hash_func", hash_func)
        cls.__init__ = init_hashes

        return cls

    if _cls:
        return _decorator(_cls)
    return _decorator


def is_watch_changes_setup(obj: ty.Any) -> bool:
    """
    :param obj: instance of any class.
    :return: `True` if `obj` class wrapped by `watch_changes`, `False` otherwise.
    """
    hashes = getattr(obj, "__wc_hashes", None)
    hash_func = getattr(obj, "__wc_hash_func", None)
    return hashes is not None and hash_func is not None


def has_changes(obj: ty.Any) -> bool:
    """
    :param obj: instance of class that wrapped by `watch_changes`.
    :return: `True` if `obj` attributes has changed, `False` otherwise.
    """
    hashes = getattr(obj, "__wc_hashes", None)
    hash_func = getattr(obj, "__wc_hash_func", None)
    if hashes is None or hash_func is None:
        raise ValueError(f"{obj} is not `watch_changes` object")

    return any(
        hashes.get(varname, None) != hash_func(str(value).encode()).hexdigest()
        for varname, value in obj.__dict__.items()
        if not varname.startswith("_")
    )


def get_changed_attributes(obj: ty.Any) -> tuple[str, ...]:
    """
    :param obj: instance of class that wrapped by `watch_changes`.
    :returns: names of attributes that have changed.
    """
    hashes = getattr(obj, "__wc_hashes", None)
    hash_func = getattr(obj, "__wc_hash_func", None)
    if hashes is None or hash_func is None:
        raise ValueError(f"{obj} is not `watch_changes` object")

    return tuple(
        varname
        for varname, value in obj.__dict__.items()
        if (
            not varname.startswith("_")
            and hashes.get(varname, None)
            != hash_func(str(value).encode()).hexdigest()
        )
    )


class Construct[T](ty.Protocol):
    def __call__(self, *args, **kwargs) -> T: ...


class DTConstruct[T](ty.Protocol):
    def fromisoformat(self, value: str) -> T: ...


def is_dt_type(obj: ty.Any) -> ty.TypeGuard[DTConstruct]:
    return obj in {datetime, date, time}


def convert_type[T](obj: ty.Any, python_type: Construct[T]) -> T:
    if is_dt_type(python_type):
        return python_type.fromisoformat(obj)
    if dataclasses.is_dataclass(python_type):
        if type(obj) is dict:
            return python_type(**obj)
        return python_type(*obj)
    return python_type(obj)



import typing as ty

from .models import Field
from .operators import ContainedCmpOperator, IsNullCmpOperator


def contains[T](
    field: Field[T], collection: list[T] | tuple[T, ...]
) -> ContainedCmpOperator:
    if not isinstance(field, Field):
        raise ValueError("first argument must be field of registered model")
    return field.contained(collection)


def is_null(field: Field) -> IsNullCmpOperator:
    if not isinstance(field, Field):
        raise ValueError("first argument must be field of registered model")
    return field.is_null()


def group_by[FT, T: ty.Any | tuple](
    field: Field[FT], objs: list[T]
) -> dict[FT, list[T]]:
    """
    Groups objects by field.
    :param field: Field for a group.
    :param objs: Model instances.
    :returns: Dict with keys - values of field and values - list of objects.
    """
    if not isinstance(field, Field):
        raise ValueError("first argument must be field of registered model")
    result: dict[FT, list[T]] = {}
    i: int | None = None
    if isinstance(next(iter(objs), None), tuple):
        i = next(
            filter(
                lambda x: field.model_name == objs[0][x].__class__.__name__,
                range(len(objs[0])),
            )
        )
    for obj in objs:
        if (
            value := getattr(obj if i is None else obj[i], field.name)
        ) not in result:
            result[value] = []
        result[value].append(obj)
    return result
